"""
Entrypoint to the discovery project
Reads a filesystem and tries to make some analysis based off of it
"""
import itertools

import yaml
import logging.config

# set up local logging before importing local libs
# TODO: do this a better way
if __name__ == "__main__":
    with open('logging_conf.yaml', 'r') as f:
        config = yaml.safe_load(f.read())
        logging.config.dictConfig(config)

    logger = logging.getLogger(__name__)

from discovery.utils.metadata.metadata import construct_metadata_from_file_descriptor
from discovery.utils.metadata.metadata_json_handler import write_metadata_to_json
from discovery.utils.decorators.type_enforcer import type_enforcer
from discovery.utils.dataframe_matcher import DataFrameMatcher
from discovery.utils.file_handler import FileHandler
from discovery.utils.decorators.persist_execution import persistence
from discovery.utils.visualizer import Visualizer


class DiscoveryClient:
    def __init__(self, discovery_config: dict):
        self.config = discovery_config
        self.file_handler = FileHandler()
        self.dataframe_file_metadata_pairs = []
        self.visualiser = Visualizer()
        self.dataframe_matcher = DataFrameMatcher(
            DataFrameMatcher.match_name_wordnet,
            DataFrameMatcher.match_data_pearson_coefficient
        )

    def create_visual(self, pathname):
        """
        Build a visual based on stored metadata
        """

        # TODO: find a fancy python way to do this
        metadata = []
        for dataframe, metadatum in self.dataframe_file_metadata_pairs:
            metadata.append(metadatum)
        self.visualiser.draw(metadata, pathname)

    def reconstruct_metadata(self):
        """
        Builds metadata based on the currently loaded files
        """
        file_metadata = []
        for path, file_descriptor in self.file_handler.loaded_files.items():
            metadatum = construct_metadata_from_file_descriptor(file_descriptor)
            file_metadata.append((file_descriptor["dataframe"], metadatum))
        self.dataframe_file_metadata_pairs = file_metadata

    def construct_relationships(self):
        for pair in itertools.combinations(self.dataframe_file_metadata_pairs, 2):
            self._match_metadata(pair[0][0], pair[0][1], pair[1][0], pair[1][1])

    def _match_metadata(self, reference_dataframe, reference_metadatum, subject_dataframe, subject_metadatum):
        """
        Run the dataframe matcher with the two given dataframes
        Update the metadata to reflect the changes
        """
        results = self.dataframe_matcher.match_dataframes(reference_dataframe, subject_dataframe)
        for relationship in results:
            column = next((x for x in reference_metadatum.columns if x.name == relationship['column_a']), None)
            if column is not None:
                column.add_relationship(
                    relationship['data_confidence'], subject_metadatum.hash, relationship['column_b']
                )

    def get_loaded_files(self):
        return self.file_handler.loaded_files

    @type_enforcer
    def add_files(self, path: str, reconstruct=False):
        self.file_handler.scan_filesystem(path)
        if reconstruct:
            self.reconstruct_metadata()

    @persistence
    @type_enforcer
    def add_file(self, path: str, reconstruct=False):
        self.file_handler.load_file(path)
        if reconstruct:
            self.reconstruct_metadata()


if __name__ == "__main__":
    # locally test the mock filesystem
    launch_config = yaml.safe_load(open("launch_config.yaml"))
    discovery_instance = DiscoveryClient(launch_config)

    from utils.datagen import FakeDataGen

    fake_data = FakeDataGen()
    fake_files = fake_data.build_df_to_file(1000, "matcher_test", index_type="categoric", continuous_data=5,
                                            file_spread=2)
    discovery_instance.add_file(fake_files[0])
    discovery_instance.add_file(fake_files[1])

    discovery_instance.reconstruct_metadata()
    discovery_instance.construct_relationships()

    for dataframe, metadata in discovery_instance.dataframe_file_metadata_pairs:
        write_metadata_to_json(metadata)

    discovery_instance.create_visual("test_visual")



"""
Reads a given file path into a dataframe
"""
import os

import pandas
import logging
from pandas.util import hash_pandas_object

from discovery.utils.custom_exceptions import UnsupportedFileExtension, FileNotFoundError
from discovery.utils.metadata_enums import FileExtension, FileSizeUnit

logger = logging.getLogger(__name__)


class FileHandler:
    def __init__(self):
        self.supported_extensions = {m.split('_')[-1]: getattr(self, m) for m in dir(self) if m.startswith('_handle')}
        self.ignored_extensions = [".metadata.json"]
        self.loaded_files = {}

    def scan_filesystem(self, file_path):
        """
        Loads a given file system into memory
        """
        for root, dirs, files in os.walk(file_path):
            for filename in files:
                if any([filename.endswith(extension) for extension in self.supported_extensions]):
                    self.load_file(os.path.join(root, filename))

    def load_file(self, file_path):
        """
        Loads a given file if a reader exists for its extension
        """
        extension = file_path.split('.')[-1] if '.' in file_path else 'txt'
        # file doesn't exist
        if not os.path.exists(file_path):
            raise FileNotFoundError(file_path)

        # file doesn't have a supported extension
        if extension not in self.supported_extensions:
            raise UnsupportedFileExtension(file_path)

        if self._is_file_metadata(file_path):
            return
        # Nothing in place to prevent reloading files for now
        handler = self.supported_extensions[extension]
        logger.debug(f"Loading file {file_path} using \"{handler.__name__}\" handler")
        self.loaded_files[file_path] = handler(file_path)

    def _is_file_metadata(self, path):
        return any(path.endswith(ignored_extension) for ignored_extension in self.ignored_extensions)

    @staticmethod
    def _handle_csv(file_path):
        dataframe = pandas.read_csv(file_path)
        return construct_file_descriptor(file_path, FileExtension.CSV, dataframe)

    @staticmethod
    def _handle_json(file_path):
        dataframe = pandas.read_json(file_path)
        return construct_file_descriptor(file_path, FileExtension.JSON, dataframe)

    @staticmethod
    def get_dataframe_hash(dataframe: pandas.DataFrame):
        return hash_pandas_object(dataframe).sum()

def construct_file_descriptor(file_path: str, extension: FileExtension, dataframe: pandas.DataFrame):
    size = os.stat(file_path).st_size
    file_hash = FileHandler.get_dataframe_hash(dataframe)
    return {
        "file_path": file_path,
        "extension": extension,
        "dataframe": dataframe,
        "size": (size, FileSizeUnit.BYTE),
        "file_hash": file_hash
    }


import json
from pandas.api.types import is_numeric_dtype
from pandas.api.types import is_string_dtype
import numbers

from discovery.utils.metadata.metadata import Metadata, ColMetadata, Relationship

def write_metadata_to_json(metadata: Metadata):
    dictionary_representation = get_metadata_dictionary_representation(metadata)
    json_string = json.dumps(dictionary_representation)
    write_json_to_file(dictionary_representation["file_path"], json_string)


def get_metadata_dictionary_representation(metadata: Metadata):
    dictionary_representation = {
        "file_path": normalize_and_get_filepath(metadata.file_path),
        "extension": normalize_and_get_extension(metadata.extension),
        "size": normalize_and_get_size(metadata.size),
        "hash": normalize_and_get_hash(metadata.hash),
        "columns": get_columns_dictionary_representation(metadata.columns)
    }
    return dictionary_representation


def get_columns_dictionary_representation(columns: [ColMetadata]):
    columns_list = []
    for column in columns:
        column_dict = {
            "name": normalize_and_get_column_name(column.name),
            "mean": normalize_and_get_column_mean(column.mean),
            "minimum": normalize_and_get_column_min_max(column.minimum),
            "maximum": normalize_and_get_column_min_max(column.maximum),
            "relationships": get_relationships_dictionary_representation(column.relationships)
        }
        if column.columns is not None:
            column_dict["columns"] = get_columns_dictionary_representation(column.columns)
        columns_list.append(column_dict)
    return columns_list


def get_relationships_dictionary_representation(relationships: [Relationship]):
    relationships_list = []
    for relationship in relationships:
        relationships_list.append(
            {
                "certainty": normalize_and_get_relationship_certainty(relationship.certainty),
                "target_file_hash": normalize_and_get_hash(relationship.target_file_hash),
                "target_column_name": normalize_and_get_column_name(relationship.target_column_name)
            }
        )
    return relationships_list


def write_json_to_file(path, json_string):
    filename = '{}.metadata.json'.format(path)
    with open(filename, 'w') as file:
        file.write(json_string)


def normalize_and_get_filepath(file_path):
    return file_path


def normalize_and_get_extension(extension):
    return extension.value


def normalize_and_get_size(size):
    return {
        "quantity": size[0],
        "unit": size[1].name
    }


def normalize_and_get_hash(file_hash):
    return file_hash


def normalize_and_get_column_name(name):
    return str(name)


def normalize_and_get_column_mean(mean):
    return normalize_pandas_numeric_values(mean)


def normalize_and_get_column_min_max(value):
    if is_string_dtype(value) or type(value) is str:
        return str(value)
    return normalize_pandas_numeric_values(value)


def normalize_pandas_numeric_values(value):
    if type(value) is float or type(value) is int:
        return value
    if is_numeric_dtype(value):
        return float(value)
    return None


def normalize_and_get_relationship_certainty(certainty):
    if isinstance(certainty, numbers.Number):
        return float(certainty)
    return None



"""
Takes two dataframes and tries to match them column by column
"""
import logging

# TODO: currently not working out of the box, must be run with the following statements...
# >>> import nltk
# >>> nltk.download('wordnet')
# >>> nltk.download('omw-1.4')
import pandas
import numpy
from difflib import SequenceMatcher
from Levenshtein import ratio
from nltk.corpus import wordnet
from itertools import product
from statsmodels.tsa.stattools import adfuller
from dtaidistance import dtw

logger = logging.getLogger(__name__)


class DataFrameMatcher:
    def __init__(self, name_matcher, data_matcher):
        self.name_matcher = name_matcher
        self.data_matcher = data_matcher

    def match_dataframes(self, df_a: pandas.DataFrame, df_b: pandas.DataFrame):
        """
        Returns all pairs of columns from dataframes A and B with their confidence percentages based on the column
        names, and on the column contents.
        :return:
        """

        similarities = []

        for column_a in df_a.columns:
            for column_b in df_b.columns:
                name_confidence = self.name_matcher(column_a, column_b)
                data_confidence = self.data_matcher(df_a.loc[:, column_a], df_b.loc[:, column_b])

                similarity = {
                    'column_a': column_a,
                    'column_b': column_b,
                    'name_confidence': name_confidence,
                    'data_confidence': data_confidence,
                }
                logger.debug(similarity)
                similarities.append(similarity)

        return similarities

    @staticmethod
    def is_column_stationary(series):
        """
        Checks if the data in a column is stationary using the Dickey-Fuller test
        :param series:
        :return:
        """
        result = adfuller(series)
        return result[0] < result[4]['5%']

    @staticmethod
    def match_data_identical_values(column_a, column_b):
        """
        Calculate a similarity percentage with set operations (intersection vs union) for (possible) categorical data
        :param column_a:
        :param column_b:
        :return:
        """

        categorical_similarity = len(numpy.intersect1d(column_a, column_b)) / len(
            numpy.union1d(column_a, column_b)) * 100
        return categorical_similarity

    @staticmethod
    def match_data_pearson_coefficient(column_a, column_b):
        """
        Calculate the Pearson correlation coefficient between the 2 columns if they are both numerical
        :param column_a:
        :param column_b:
        :return:
        """

        if pandas.api.types.is_numeric_dtype(column_a) and pandas.api.types.is_numeric_dtype(column_b):
            pearson_correlation_coefficient = abs(column_a.corr(column_b)) * 100
            return pearson_correlation_coefficient
        return 0

    @staticmethod
    def match_data_dynamic_time_warping(column_a, column_b):
        """
        Calculate the normalized distance measure between two numerical columns using Dynamic Time Warping
        :param column_a:
        :param column_b:
        :return:
        """

        max_distance = len(column_a) * max(column_a)

        if pandas.api.types.is_numeric_dtype(column_a) and pandas.api.types.is_numeric_dtype(column_b):
            return (max_distance - dtw.distance(column_a, column_b)) / max_distance * 100
        return 0

    @staticmethod
    def match_data_trends(column_a, column_b):
        pass

    @staticmethod
    def match_name_lcs(name_a, name_b):
        """
        Calculate the similarity of 2 column names using the Longest Contiguous Matching Subsequence
        :param name_a:
        :param name_b:
        :return:
        """

        return SequenceMatcher(None, name_a, name_b).ratio() * 100

    @staticmethod
    def match_name_levenshtein(name_a, name_b):
        """
        Calculate the similarity of 2 column names using the Levenshtein distance
        :param name_a:
        :param name_b:
        :return:
        """

        return ratio(name_a, name_b) * 100

    @staticmethod
    def match_name_wordnet(name_a, name_b):
        """
        Calculate the similarity of 2 column names using WordNet
        :param name_a:
        :param name_b:
        :return:
        """

        synset_first = wordnet.synsets(name_a)
        synset_second = wordnet.synsets(name_b)
        wordnet_ratio = 0
        if len(synset_first) > 0 and len(synset_second) > 0:
            wordnet_ratio = max(
                wordnet.wup_similarity(s1, s2) for s1, s2 in product(synset_first, synset_second)) * 100

        return wordnet_ratio



"""
Data storage in memory
"""
from typing import Union

from abc import ABC, abstractmethod
from pandas.api.types import is_numeric_dtype

from discovery.utils.metadata_enums import FileSizeUnit, FileExtension


class Relationship:
    certainty: int
    target_file_hash: int
    target_column_name: str

    def __init__(self, certainty, target_file_hash, target_column_name):
        self.certainty = certainty
        self.target_file_hash = target_file_hash
        self.target_column_name = target_column_name


class ColMetadata(ABC):
    name: str
    col_type: str
    columns: any  # [ColMetadata]
    relationships: [Relationship]

    def __init__(self, name: str, col_type: str, columns=None):
        self.name = name
        self.col_type = col_type
        self.columns = columns
        self.relationships = []

    def set_columns(self, columns):
        self.columns = columns

    def add_relationship(self, certainty, target_file_hash, target_column_name):
        self.relationships.append(Relationship(certainty, target_file_hash, target_column_name))


class NumericColMetadata(ColMetadata):
    mean: Union[float, None]
    minimum: any
    maximum: any

    def __init__(self, name: str, col_type: str, mean: Union[int, float, None], min_val, max_val, columns=None):
        ColMetadata.__init__(self, name, col_type, columns)
        self.mean = mean
        self.minimum = min_val
        self.maximum = max_val


class CategoricalColMetadata(ColMetadata):
    def __init__(self, name: str, col_type: str, columns=None):
        ColMetadata.__init__(self, name, col_type, columns)


class Metadata:
    def __init__(self, file_path: str, extension: FileExtension,
                 size: (int, FileSizeUnit), file_hash: int,
                 columns: [] = []):
        self.file_path = file_path
        self.extension = extension
        self.size = size
        self.hash = int(file_hash)
        self.columns = columns


def construct_metadata_from_file_descriptor(file_descriptor):
    metadatum = Metadata(file_descriptor["file_path"], file_descriptor["extension"],
                         file_descriptor["size"], file_descriptor["file_hash"])
    col_meta = []
    dataframe = file_descriptor["dataframe"]

    for col_name in dataframe.columns:
        column_data = construct_column(dataframe[col_name])
        col_meta.append(column_data)
    metadatum.columns = col_meta
    return metadatum


def construct_column(column):
    average, col_min, col_max = get_col_statistical_values(column)
    return NumericColMetadata(column.name, column.dtype, average, col_min, col_max)


def get_col_statistical_values(column):
    average, col_min, col_max = None, None, None

    col_min = column.min()
    col_max = column.max()

    if is_numeric_dtype(column):
        average = column.mean()
    return average, col_min, col_max



from enum import Enum


class FileSizeUnit(Enum):
    BYTE = "byte"
    KILOBYTE = "kilobyte"
    MEGABYTE = "megabyte"
    GIGABYTE = "gigabyte"


class FileExtension(Enum):
    JSON = "json"
    CSV = "csv"



from typing import Optional
import graphviz

from discovery.utils.metadata.metadata import Metadata


class SubGraphNode:
    children: []
    filesystem_full_path: str
    graph: graphviz.Digraph

    def __init__(self, filesystem_full_path, graph: Optional[graphviz.Graph] = None):
        self.children = []
        self.filesystem_full_path = filesystem_full_path
        self._set_graph(graph)

    def _set_graph(self, graph: Optional[graphviz.Graph]):
        if graph is None:
            self.graph = self._generate_graph()
        else:
            self.graph = graph

    def _generate_graph(self):
        subgraph = graphviz.Graph(name="cluster_" + self.filesystem_full_path)
        subgraph.attr(label=self.filesystem_full_path)
        subgraph.attr(color='grey')
        return subgraph

    def get_parent_path(self):
        split = self.filesystem_full_path.rsplit('/', 1)
        if 1 < len(split):
            return split[0]
        return ""

    def recursively_append_subgraphs(self):
        for child in self.children:
            child.recursively_append_subgraphs()
            self.graph.subgraph(child.graph)
        return


class Visualizer:
    engine: str

    root: SubGraphNode
    working_node: SubGraphNode
    observed_nodes: [SubGraphNode]

    def __init__(self, engine="fdp"):
        self.engine = engine
        self.observed_nodes = []
        self.root = SubGraphNode("", self._generate_root_graph())
        self.working_node = self.root
        self.observed_nodes.append(self.working_node)

    def _generate_root_graph(self):
        start_graph = graphviz.Graph(name="cluster_root", comment='File Visualization',
                                     node_attr={'shape': 'plaintext'})
        start_graph.attr(label="")
        start_graph.engine = self.engine
        return start_graph

    def draw(self, metadata: [Metadata], filename: str):
        self._draw_metadata(metadata)
        # self.draw_relationships(metadata)
        self._finalize_result_graph(filename)

    def _draw_metadata(self, metadata):
        for datum in metadata:
            self.working_node = self._determine_working_node(datum.file_path)
            self._draw_metadatum(datum)

    # Don't use it, not implemented properly
    def _draw_relationships(self, metadata):
        for metadatum in metadata:
            for column in metadatum.columns:
                for relationship in column.relationships:
                    self.root.graph.edge(f"{metadatum.hash}:{column.name}",
                                         f"{relationship.target_file_hash}:{relationship.target_column_name}",
                                         str(relationship.certainty)
                                         )

    def _finalize_result_graph(self, output_filename):
        self.root.recursively_append_subgraphs()
        self.root.graph.view(output_filename)

    def _draw_metadatum(self, metadatum: Metadata):
        columns = self._draw_table_columns(metadatum)
        filled_table = self._draw_filled_metadatum_table(metadatum.file_path, columns)
        self.working_node.graph.node(str(metadatum.hash), filled_table)

    # TODO: make it more generic
    def _draw_table_columns(self, metadatum):
        col_rows: str = ""
        for column in metadatum.columns:
            col_rows += '<TR><TD>{}</TD> <TD>{}</TD> <TD>{}</TD> <TD>{}</TD> <TD>{}</TD></TR>' \
                .format(column.name, column.col_type,
                        (column.mean if column.mean is not None else "NA"),
                        column.minimum, column.maximum)
        return col_rows

    # TODO: find a more generic approach
    def _draw_filled_metadatum_table(self, filename: str, col_strings: str):
        return f'''<
            <TABLE BORDER="0" CELLBORDER="1" CELLSPACING="0">
              <TR>
                <TD COLSPAN="2" BGCOLOR="darkgrey">{filename}</TD>
              </TR>
              <TR>
                <TD BGCOLOR="lightgray">Name</TD>
                <TD BGCOLOR="lightgray">Type</TD>
                <TD BGCOLOR="lightgray">Average</TD>
                <TD BGCOLOR="lightgray">Lowest</TD>
                <TD BGCOLOR="lightgray">Highest</TD>

              </TR>
              {col_strings}
            </TABLE>>'''

    def _determine_working_node(self, filesystem_full_path: str):
        determined_node = None

        already_visited_node = self._get_node_if_path_was_already_observed(filesystem_full_path)
        if already_visited_node is not None:
            determined_node = already_visited_node
        else:
            determined_node = SubGraphNode(filesystem_full_path)
            self._add_all_previously_unobserved_parents_to_observed_nodes(determined_node)
        return determined_node

    def _add_all_previously_unobserved_parents_to_observed_nodes(self, parameter_child_node: SubGraphNode):
        working_child_node = parameter_child_node
        while True:
            visited_parent = self._get_node_if_path_was_already_observed(working_child_node.filesystem_full_path)
            if visited_parent is None:
                new_parent = SubGraphNode(working_child_node.get_parent_path())
                new_parent.children.append(working_child_node)
                # setup next loop
                working_child_node = new_parent
            else:
                visited_parent.children.append(working_child_node)
                break

    def _get_node_if_path_was_already_observed(self, filesystem_full_path: str):
        return next(iter([node for node in self.observed_nodes if node.filesystem_full_path == filesystem_full_path]),
                    None)




"""
Ensure that a custom exception doesn't terminate the process
"""
import logging

from discovery.utils.custom_exceptions import CustomException


def persistence(function):
    """
    if the method is a custom exception, ignore it
    """
    logger = logging.getLogger(__name__)

    def wrapper_function(*args, **kwargs):
        try:
            function(*args, **kwargs)
        except CustomException as e:
            logger.error(e)

    return wrapper_function
